unit IBQODB;

interface
uses classes,sysutils,QuickODB,QuickRTTI, IBDatabase, IBSQL, Db;

type

TIBQODB = class(TQODB)
  private
    fclasses:TCollection;
    fObjects:TStringlist;
    fdb:TIBDatabase;
    ftran:TIBTransaction;
    function   ClassExists(classname:String):boolean;
    procedure  CreateClassDefinition (Def:TQODBClassDefinition);
    procedure  LoadClassDefinitions;
  protected       
  public
    constructor create;override;
   procedure   RegisterClass (RegisteredClass:TClass;Prototype:TObject); override;
 {  function    CreateObject  (RegisteredClass:TClass;Prototype:TObject):TObjectID;override;
    procedure   DeleteObject  (OID:TObjectID);override;
    procedure   LoadObject    (OID:TObjectID;Obj:TObject);override;
    procedure   SaveObject    (OID:TObjectID;Obj:TObject);override;}
  published
    procedure   Open(database,uid,pass:String);
    procedure   Close;
  end;

{
TABLES:

_CLASSES
   NAME
   PROPERTYNAME
   PROPERTYTYPE
   ISOBJECT

Each class gets a table created dynamically for it. Collection properties not yet supported... sub objects yes.
Main Objects have an OID, and subobjects reference this. properties span several tables.. so
that a TUser might store username and password, while TCustomer=(TUser) might have address info... both
with the same OID.

Objects should map to Delphi objects once Loaded from the database... object references should be checked.
object refs should also be OID's in the db, but translated to local objects. When reading, if no local copy exists, then we need
to create/load and object of that type.
}

implementation


constructor TIBQODB.create;
begin
  inherited create;
  fdb:=TIBDatabase.create(nil);
  fdb.loginprompt:=false;
  ftran:=TIBTransaction.create(nil);
  ftran.defaultdatabase:=fdb;
  fdb.defaulttransaction:=ftran;
end;

procedure  TIBQODB.Open(database,uid,pass:String);
begin
  fdb.close;
  fdb.databasename:=database;
  fdb.params.clear;
  fdb.params.add('user_name='+uid);
  fdb.params.add('password='+pass);
  fdb.open;
  ftran.starttransaction;
  LoadClassDefinitions;
end;

procedure  TIBQODB.Close;
begin
  ftran.commit;
  fdb.close;
end;

function TIBQODB.ClassExists(classname:String):boolean;
var Find:TIBSQL;
begin
  Find:=TIBSQL.create(nil);
  if not(ftran.active) then ftran.starttransaction;
  Find.database:=fdb;
  Find.transaction:=ftran;
  Find.sql.text:='Select count(distinct Classname) from QODB_Classes where classname=:classname';
  Find.ParamByName('classname').asstring:=classname;
  FInd.execquery;
  if find.fields[0].value=0 then
     result:=false
     else
     result:=true;
  find.close;
  find.free;
  ftran.commit;
end;

procedure TIBQODB.RegisterClass (RegisteredClass:TClass;Prototype:TObject);
var thisclass:String; Def:TQODBClassDefinition ;
begin
  inherited Registerclass(RegisteredClass,Prototype);
  thisclass:=RegisteredClass.classname;
  def:=FindClass(thisclass);
  if not(ClassExists(thisclass)) then
   begin
     CreateClassDefinition (Def);
   end;
end;


procedure TIBQODB.CreateClassDefinition (Def:TQODBClassDefinition);
var PutClass,MakeTable:TIBSQL; i,max:integer;op:TQODBProperty;  optype:STring;
begin
    if GetCLass(Def.Ancestor)<>TPersistent then
       CreateClassDefinition(FindClass(Def.ancestor));

 if not(classexists(Def.Name)) then
  begin
    PutClass:=TIBSQL.create(nil);
    Putclass.database:=fdb;
    PutClass.transaction:=ftran;
    if not(ftran.active) then ftran.starttransaction;
    try
    putclass.sql.text:='INSERT INTO QODB_CLASSES (CLASSNAME,PROPERTYNAME,PROPERTYTYPE,ISOBJECT) VALUES (:CLASSNAME,:PROPERTYNAME,:PROPERTYTYPE,:ISOBJECT)';
    max:=Def.Properties.count-1;
    for i:= 0 to max do
      begin
        op:=TQODBProperty(def.properties.items[i]);
        putclass.parambyname('CLASSNAME').asstring:= def.name;
        putclass.parambyname('PROPERTYNAME').asstring:= op.Name ;
        putclass.parambyname('PROPERTYTYPE').asstring:= op.PropertyType ;
        if op.isobject then
          putclass.parambyname('ISOBJECT').asstring:='Y'
          else
          putclass.parambyname('ISOBJECT').asstring:='N';
        putclass.execquery;
      end;
    PutClass.close;

    {Create the storage table}
    MakeTable:=TIBSQL.create(nil);
    MakeTable.database:=fdb;
    MakeTable.transaction:=ftran;
    MakeTable.sql.add('CREATE TABLE '+Def.name+' (');
    MakeTable.sql.add('  OID varchar(40) not null,');
     for i:= 0 to max do
      begin
           op:=TQODBProperty(def.properties.items[i]);
         optype:=UPPERCASE(op.PropertyType) ;
         if optype='STRING' then optype:='VARCHAR(255)';
         if optype='TDATETIME' then optype:='TIMESTAMP';
         if optype='CHAR' then optype:='CHAR(1)';
         if optype='BOOLEAN' then optype:='CHAR(1)';
         if op.IsObject then optype:='VARCHAR(40)';
         MakeTable.sql.add (op.Name+'     '+optype+'') ;
         if i<max then MakeTable.sql.text:=MakeTable.sql.text+',';
      end;
    MakeTable.sql.add (');');

    MakeTable.ExecQuery ;
    MakeTable.close ;
    MakeTable.free;
    Putclass.free;
    ftran.commit;
    except on E:Exception do
      ftran.rollback;
    end;
  end;
end;

procedure  TIBQODB.LoadClassDefinitions;
var Find,GetClass:TIBSQL; Found:tstringlist; cd:TQODBClassdefinition; n,ot,io:String;
begin
  Find:=TIBSQL.create(nil);
  GetCLass:=TIBSQL.create(nil);
  if not(ftran.active) then ftran.starttransaction;
  Find.database:=fdb;
  GetClass.database:=fdb;
  Find.transaction:=ftran;
  GetClass.transaction:=ftran;
  GetClass.sql.text:='Select * from QODB_Classes where classname=:classname';
  Find.sql.text:='Select distinct Classname from QODB_Classes where classname=:classname';
  Find.ParamByName('classname').asstring:=classname;
  FInd.execquery;
  while not(Find.eof) do
  begin
     cd:=TQODBClassdefinition(findclass(Find.fields[0].asstring));
     if cd=nil then
     begin
     GetCLass.ParamByName('classname').asstring:=Find.fields[0].asstring;
     GetCLass.execquery;
     cd:= TQODBClassdefinition(ClassDefinitions.add);
     while not (getclass.eof) do
      begin
        n := GetCLass.fieldbyname('PropertyName').asstring;
        ot:= GetCLass.fieldbyname('PropertyType').asstring;
        io:= GetCLass.fieldbyname('IsObject').asstring;
        if io='Y' then
           cd.AddProperty(n,ot,true)
           else
           cd.AddProperty(n,ot,false);
      end;
     {clean up}
     GetClass.close;
     end;
    Find.Next ;
  end;
  find.close;
  find.free;
  getclass.free;
  ftran.commit;
end;


end.
