unit QODB;

interface
uses classes,sysutils,Quickrtti,guids;

type

TObjectID = string;

TQODBObject = class(TCollectionItem)
  private
     fobj:TObject;
     foid,fown:TObjectID;
     fclass:String;
     fprops:tstringlist;
  public
    constructor create(Collection:TCollection);override;
    destructor  destroy;override;
  published
     property ObjectClassName:String  read fclass write fclass;
     property OwnerOID:TObjectID read fown write fown;
     property OID:TObjectID read foid write foid;
     property Data:tstringlist read fprops;
  end;

TQODBProperty = class(TCollectionItem)
   private
      fname,ftype:String; fisobject:boolean;   
   published
      property Name:String read fname write fname;
      property ObjectType:String read ftype write ftype;
      property IsObject:boolean read fisObject write fisobject;
   end;

TQODB = class ; {forward declaration}

TQODBClass = class(TCollectionItem)
   private
    fname:String; fDB:TQODB;
    fancestor:String; fclass:TClass;
    fprops,fobjects:TCollection;
   public
    constructor create(Collection:TCollection);override;
    destructor  destroy;override;
    function  AddProperty(Name,ObjectType:String;IsObject:boolean):TQODBProperty;
    function  GetObject(OID:TObjectID):TQODBObject;
    function  AddObject(OID:TObjectID):TQODBObject;
    procedure DeleteObject (OID:TObjectID);
    property DB:TQODB read fdb write fdb;
   published 
    property Name:String read fname write fname;
    property Ancestor:String read fancestor write fancestor;
    property ObjectClass:TClass read fclass write fclass;
   end;


TQODB = class(TPersistent)
  protected
    function  GetClass(classname:String):TQODBClass;virtual;abstract;
    function  GetObject (OID:TObjectID):TQODBObject;virtual;abstract;
  public
    constructor create;virtual;abstract;
    procedure   RegisterClass (RegisteredClass:TClass;Prototype:TObject); virtual;abstract;
    function    CreateObject  (RegisteredClass:TClass;Prototype:TObject):TObjectID;virtual;abstract;
    procedure   DeleteObject (OID:TObjectID);virtual; abstract;
    procedure   LoadObject (OID:TObjectID;Obj:TObject);virtual; abstract;
    procedure   SaveObject (OID:TObjectID;Obj:TObject);virtual; abstract;
  end;

TMemQODB = class(TQODB)
  private
    fObjects:TList;
    fclasses:TCollection;
    fproperties:Tstringlist;
  protected
    property CLasses:TCollection read fclasses write fclasses;
    property Objects:TList read fobjects write fobjects;
    function  GetClass(classname:String):TQODBClass; override;
    function  GetObject (OID:TObjectID):TQODBObject; override;
    procedure SaveClassObject (OID:TObjectID;Obj:TObject;Prototype:TClass;Owned:Boolean);
    procedure LoadClassObject (OID:TObjectID;Obj:TObject;Prototype:TClass);
    procedure InsertObject (oid:TObjectID;Classname:String); 
  public
    constructor create;override;
    destructor  destroy;override;
    procedure   RegisterClass (RegisteredClass:TClass;Prototype:TObject); override;
    function    CreateObject  (RegisteredClass:TClass;Prototype:TObject):TObjectID;override;
    procedure   DeleteObject (OID:TObjectID);override;
    procedure   LoadObject (OID:TObjectID;Obj:TObject);override;
    procedure   SaveObject (OID:TObjectID;Obj:TObject);override;
  end;


  procedure ClassPropsList (Objectclass:TClass;s:TStrings);

implementation


constructor TQODBObject.create (Collection:TCollection);
begin
  inherited create(Collection);
  fprops:=tstringlist.create;
end;

destructor TQODBObject.destroy;
begin
  try
    fprops.free;
  finally
    inherited destroy;
  end;
end;

procedure  TMemQODB.InsertObject (OID:TObjectID;classname:String);
var oClass:TQODBClass;  factory:TObject;
begin
 {Go find that class!}
 oClass:=GetCLass(classname);
 oClass.AddObject (oid);
 fobjects.Add(oCLass.GetObject(OID));
 factory:=oClass.ObjectClass.NewInstance ;
 factory.create;
 SaveClassObject(oid,factory,oClass.ObjectClass,true);
 factory.free;
 {Ok..base class added.} 
end;

function  TMemQODB.CreateObject (RegisteredClass:TClass;Prototype:TObject):TObjectID;
var oClass:TQODBClass; oid:TObjectID;
begin
 {Go find that class!}
 oClass:=GetCLass(RegisteredClass.classname);
 oid:=guids.createobjectid;
 oClass.AddObject (oid);
 fobjects.Add(oCLass.GetObject(OID));
 SaveClassObject(oid,Prototype,oClass.ObjectClass,true); 
 {Ok..base class added.}
 result:=oid;
end;

procedure  TMemQODB.DeleteObject (OID:TObjectID);
var max,i:integer; oClass:TQODBClass;  o:TQODBObject;
begin

  max:=self.fobjects.Count -1;
  for i:= max downto 0 do
   begin
      o:=TQODBObject(fobjects.items[i]);
     if assigned(o) then
     begin
      oClass:= GetCLass(o.objectclassname);
      if o.OID = oid then
         begin
          oCLass.DeleteObject(o.oid);
          fobjects.delete(i);
         end;
     end;
   end;

  max:=self.fobjects.Count -1;
  for i:= max downto 0 do
   begin
      o:=TQODBObject(fobjects.items[i]);
     if assigned(o) then
     begin
      oClass:=GetCLass(o.objectclassname);
      if o.OwnerOID = oid then
        begin
         DeleteObject(o.oid);
        end;
     end;
   end;



end;

function  TMemQODB.GetObject (OID:TObjectID):TQODBObject;
var max,i:integer; oClass:TQODBClass;  o:TQODBObject;
begin
  max:=self.fobjects.Count -1;
  for i:= max downto 0 do
   begin
      o:=TQODBObject(fobjects.items[i]);
      if o.OID = oid then result:=o;
   end;
end;


procedure BuildHierarchypropslist (classname:String;Objectclass:TClass; s:TStrings);
var hold,check:Tstringlist; Q:TRTTIEnabler; o:TPersistent;
    i,max,j,jmax,idx:integer; bfound:boolean;thisprop:String;
begin

  if ObjectClass<>TObject then
   begin
  hold:=Tstringlist.create;
  BuildHierarchypropslist(classname,objectclass.ClassParent,s);
  o:=TPersistent(ObjectClass.NewInstance);
  o.create;
  Q:=TRTTIEnabler.create;
  Q.rttiobject:=o;
  max:= Q.propertyCount -1;
  for i:= 0 to max do
   begin
   jmax:=s.count-1 ;
   bfound:=false;
   idx:=-1;
   thisprop :=Q.propertynames(i)+':'+Q.propertyVarTypes(i);
   for j:=0 to jmax do
    begin
      check:=TStringlist(s.objects[j]);
      if check.indexof(thisprop)>-1 then
         begin
           bfound:=true;
           idx:=check.indexof(thisprop);
         end;
    end;
   if  not(bfound)  then
     begin
       hold.Add (thisprop);
     end;
   end;
  Q.free;
  o.free;
  s.addobject (objectclass.classname,hold) ;
  end;

end;


procedure ClassPropsList (Objectclass:TClass; s:TStrings);
var hold:Tstringlist; max,i:integer;
begin
  hold:=Tstringlist.create;
  BuildHierarchypropslist(Objectclass.classname,ObjectClass,hold);
  max:= hold.count-1;
  for i:=0 to max do
   begin
    if hold[i]=Objectclass.classname then s.text:=tstringlist(hold.objects[i]).text;
   end;
  hold.free;
end;


procedure TMemQODB.RegisterClass (RegisteredClass:TClass;Prototype:TObject);
var o:TPersistent; obj:TObject; Q:TRTTIEnabler; ancestor:TClass; theseprops,anc,properties:tstringlist;
    i,max,j,jmax :integer;  bFOund:boolean; thisprop:String; thisclass: TQODBClass;
    ClassProtoType:TClass;
begin
  obj:=Prototype;
 if obj is TPersistent then
  begin
    o:=TPersistent(RegisteredClass.newinstance);
    o.create;
    theseprops:=tstringlist.create;
    Q:=TRTTIEnabler.create;
    Q.RTTIObject := o;
    Q.TagName := '';
    max:=Q.propertyCount -1;
    ancestor:=registeredclass.ClassParent ;
    if (ancestor<>TPersistent) then RegisterClass (ancestor ,ProtoType);
     thisclass:= TQODBClass(fclasses.add);
     thisclass.ObjectClass := registeredclass;
     thisclass.db := self;
     thisclass.Name := registeredclass.classname;
     thisclass.Ancestor := ancestor.ClassName ;
    for i:= 0 to max do
     begin

        thisprop:=Q.propertynames(i)+':'+Q.propertyVarTypes(i);
        jmax:= fProperties.count-1;
        bfound:=false;
        for j:= 0 to jmax do
         begin
          anc:= Tstringlist(fProperties.objects[j]);
          if anc.indexof(thisprop)>-1 then
               bfound:=true;
         end;
         if not(bFound) then
           begin
           theseprops.add (thisprop) ; {only add if NOT in ancestors}
           thisclass.AddProperty( Q.propertynames(i),Q.propertyVarTypes(i),(ord(Q.propertyTypes (i))=7) );
           end;
     end;
    o.free;
    fProperties.addobject(registeredclass.classname,theseprops);
    q.free;
  end;
  if registeredclass = prototype.ClassType then obj.free;
end;

procedure TMemQODB.LoadObject (OID:TObjectID;Obj:TObject);
begin
   LoadClassObject(oid,obj,obj.ClassType );
end;

procedure TMemQODB.SaveObject (OID:TObjectID;Obj:TObject);
begin
   SaveClassObject(oid,obj,obj.ClassType ,false);
end;


procedure TMemQODB.SaveClassObject (OID:TObjectID;Obj:TObject;Prototype:TClass;Owned:boolean);
var o:TPersistent;   Q:TRTTIEnabler; ancestor:TClass; theseprops,anc,properties:tstringlist;
    i,max,j,jmax :integer;  bFOund:boolean; thisprop,thistype:String; thisclass: TQODBClass; storedobj,ownedObj:TQODBObject;
    ObjectPointer:TObjectID;
begin
 if obj is TPersistent then
  if prototype<>TPersistent then
  begin
    o:=TPersistent(obj); 
    Q:=TRTTIEnabler.create;
    Q.RTTIObject := o;
    Q.TagName := '';
    max:=Q.propertyCount -1;
    ancestor:=prototype.ClassParent  ;
    SaveClassObject (oid,obj,ancestor,Owned); {Save recursively.}
    thisclass:= GetClass(prototype.ClassName);
    theseprops:=tstringlist(fProperties.objects[fProperties.indexof(Prototype.classname)]);
    storedobj:= thisclass.GetObject(OID);
    for i:= 0 to max do
    if ord(Q.propertyTypes(i))<>7 then  //if Q.propertyTypes(i)<>tkClass then
     begin
        thisprop:=Q.propertynames(i);
        thistype:=Q.propertyVarTypes (i);
        if theseprops.indexof(thisprop+':'+thistype)>-1 then storedobj.Data.Values [thisprop] := Q.Value[thisprop];
     end else
      begin
        thisprop:=Q.propertynames(i);   
        thistype:=Q.propertyVarTypes (i);
        ObjectPointer:=storedobj.Data.Values [thisprop];
        if ((ObjectPointer='') and (assigned(Q.ChildObject (thisprop)))) then
         if theseprops.indexof(thisprop+':'+thistype)> -1 then
          begin
            if assigned(Q.ChildObject (thisprop)) then
             begin
                ObjectPointer:= CreateObject(Q.ChildObject(thisprop).classtype, Q.ChildObject (thisprop));
                fobjects.Add(GetObject(ObjectPointer));
                storedobj.Data.Values [thisprop]:=ObjectPointer;
                SaveObject(ObjectPointer,Q.ChildObject (thisprop));
                if Owned then
                  begin
                    OwnedObj:= GetObject(ObjectPointer);
                    OwnedObj.OwnerOID := oid;
                  end;
             end;
          end;
      end;
    q.free;
  end;
end;


procedure TMemQODB.LoadClassObject (OID:TObjectID;Obj:TObject;Prototype:TClass);
var o:TPersistent;   Q:TRTTIEnabler; ancestor:TClass; theseprops,anc,properties:tstringlist;
    i,max,j,jmax :integer;  bFOund:boolean; thisprop,thistype:String; thisclass: TQODBClass; storedobj,ownedobj:TQODBObject;
    ObjectPointer:TObjectID;
begin
 if obj is TPersistent then
   if prototype<>TPersistent then
  begin
    o:=TPersistent(obj); 
    Q:=TRTTIEnabler.create;
    Q.RTTIObject := o;
    Q.TagName := '';
    max:=Q.propertyCount -1;
    ancestor:=prototype.ClassParent  ;
    LoadClassObject (oid,obj,ancestor); {Save recursively.}
    thisclass:= GetClass(prototype.ClassName);
    theseprops:=tstringlist(fProperties.objects[fProperties.indexof(Prototype.classname)]);
    storedobj:= thisclass.GetObject(OID);
  for i:= 0 to max do
   if ord(Q.propertyTypes(i))<>7 then  //if Q.propertyTypes(i)<>tkClass then
     begin
        thisprop:=Q.propertynames(i);
        thistype:=Q.propertyVarTypes (i);
        if theseprops.indexof(thisprop+':'+thistype)> -1 then Q.Value[thisprop]:=storedobj.Data.Values [thisprop];
     end else
      begin
        thisprop:=Q.propertynames(i);
        thistype:=Q.propertyVarTypes (i);
        ObjectPointer:=storedobj.Data.Values [thisprop];
        if theseprops.indexof(thisprop+':'+thistype)> -1 then
        if ObjectPointer<>'' then
          begin
             ownedobj:=GetObject(ObjectPointer);
             if ownedobj.OwnerOID = oid then
               begin
                LoadObject(ObjectPointer,Q.ChildObject (thisprop));
               end;
          end;
      end;
    q.free;
  end;
end;


constructor TMemQODB.create;
begin
  fObjects:=tList.create;
  fClasses:=tcollection.create(TQODBClass);
  fproperties:=tstringlist.create;
end;

destructor TMemQODB.destroy;
begin
  try
  fObjects.free;
  fproperties.free;
  finally
   inherited destroy;
  end;
end;

function TMemQODB.GetClass(classname:String):TQODBClass;
var i,max,idx:integer;found:boolean; hold:TQODBClass;
begin
found:=false;
i:=0;
max:=fclasses.count-1;
while (i<=max) do
begin
 hold:=TQODBClass(fclasses.items[i]);
 if hold.Name = classname then
  begin
    idx:=i;
    found:=true;
  end;
 inc(i);
end;
if found then
 begin
    result:=TQODBClass(fclasses.items[idx]);
 end;
end;

constructor TQODBClass.create(Collection:TCollection);
begin
 inherited create(COllection);
 fprops:=TCollection.create(TQODBProperty);
 fobjects:=TCollection.create(TQODBObject);
end;

function TQODBClass.GetObject(OID:TObjectID):TQODBObject;
var i,max,idx:integer;found:boolean; hold:TQODBObject;
begin
found:=false;
i:=0;
max:=fobjects.count-1;
while (i<=max) do
begin
 hold:=TQODBObject(fobjects.items[i]);
 if hold.oid=oid then
  begin
    idx:=i;
    found:=true;
  end;
 inc(i);
end;
if found then
 begin
    result:=TQODBObject(fobjects.items[idx]);
 end;
end;

function TQODBClass.AddObject(oid:TObjectID):TQODBObject;
var anc:TQODBClass;
begin
  result:= TQODBObject(fobjects.add);
  result.ObjectClassName := self.Name ;
  result.oid:=oid ;
  if ((self.Ancestor <>'TPersistent') and (self.ancestor<>'')) then
     begin
        anc:= fdb.GetClass(self.Ancestor);
        anc.AddObject (oid);
     end;
end;

procedure TQODBClass.DeleteObject (OID:TObjectID);
var anc:TQODBClass;i,max:integer; o:TQODBObject;
begin
  max:=fobjects.count-1;
  for i:= max downto 0 do
   begin
      o:=TQODBObject(fobjects.items[i]);
      if o.OID = oid then fobjects.Delete (i);
   end;
end;

destructor  TQODBClass.destroy;
begin
 try
   fprops.free;
 finally
   inherited destroy;
 end;
end;

function TQODBClass.AddProperty(Name,ObjectType:String;IsObject:boolean):TQODBProperty;
begin
  result:=TQODBProperty(fprops.add);
  result.name:=name;
  result.objecttype:=ObjectType;
  result.isobject:=isobject;
end;




end.
