unit QuickODB;

interface
uses classes,sysutils,Quickrtti;

type
TObjectID = string;

TQODBProperty = class(TCollectionItem)
   private
      fname,ftype:String; fisobject:boolean;   
   published
      property Name:String read fname write fname;
      property PropertyType:String read ftype write ftype;
      property IsObject:boolean read fisObject write fisobject;
   end;

TQODB=class;   

TQODBClassDefinition = class(TCollectionItem)
   private
    fname:String; fDB:TQODB;
    fancestor:String; fclass:TClass;
    fprops:TCollection;
   public
    constructor create(Collection:TCollection);override;
    destructor  destroy;override;
    function    AddProperty(Name,ObjectType:String;IsObject:boolean):TQODBProperty;
   published
    property Name:String read fname write fname;
    property Ancestor:String read fancestor write fancestor;
    property ObjectClass:TClass read fclass write fclass;
    property Properties:TCollection read fprops;
   end;

TQODB = class(TPersistent)
  private
    fclasses:TCollection;
    fObjects:TStringlist;
  protected
    function    FindObject    (OID:TObjectID):TObject; {get obj ref from OID}
    function    FindOID       (Obj:TObject):TObjectID; {get objectID from obj ref}
    function    FindClass     (classname:String):TQODBClassDefinition;
    procedure   AddObjectRef   (OID:String;Obj:TObject);
    procedure   RemoveRef      (OID:String);
  public
    constructor create; virtual; 
    procedure   RegisterClass (RegisteredClass:TClass;Prototype:TObject); virtual;
    function    CreateObject  (RegisteredClass:TClass;Prototype:TObject):TObjectID;virtual;abstract;
    procedure   DeleteObject  (OID:TObjectID);virtual; abstract;
    procedure   LoadObject    (OID:TObjectID;Obj:TObject);virtual; abstract;
    procedure   SaveObject    (OID:TObjectID;Obj:TObject);virtual; abstract;
  published
    property    ClassDefinitions:TCollection read fclasses;
  end;

 procedure ClassPropsList (Objectclass:TClass;s:TStrings);

implementation

constructor TQODB.create;
begin
  fobjects:=tstringlist.create;
  fclasses:=tcollection.create(TQODBClassDefinition);
end;

procedure   TQODB.AddObjectRef   (OID:String;Obj:TObject);
begin
  fobjects.addobject(OID,Obj);
end;

procedure   TQODB.RemoveRef      (OID:String);
var idx:integer;
begin
  idx:= fobjects.indexof(OID);
  fobjects.delete(idx);
end;

procedure BuildHierarchypropslist (classname:String;Objectclass:TClass; s:TStrings);
var hold,check:Tstringlist; Q:TRTTIEnabler; o:TPersistent;
    i,max,j,jmax,idx:integer; bfound:boolean;thisprop:String;
begin
  {Called by ClassPropsList, creates a tstringlist of tstringlists}
  if ObjectClass<>TObject then
   begin
  hold:=Tstringlist.create;
  BuildHierarchypropslist(classname,objectclass.ClassParent,s);
  o:=TPersistent(ObjectClass.NewInstance);
  o.create;
  Q:=TRTTIEnabler.create;
  Q.rttiobject:=o;
  max:= Q.propertyCount -1;
  for i:= 0 to max do
   begin
   jmax:=s.count-1 ;
   bfound:=false;
   idx:=-1;
   if ord(Q.propertyTypes(i))=7 then //tkclass
      thisprop :=Q.propertynames(i)+'=*'+Q.propertyVarTypes(i)
    else
      thisprop :=Q.propertynames(i)+'='+Q.propertyVarTypes(i);
   for j:=0 to jmax do
    begin
      check:=TStringlist(s.objects[j]);
      if check.indexof(thisprop)>-1 then
         begin
           bfound:=true;
           idx:=check.indexof(thisprop);
         end;
    end;
   if  not(bfound)  then
     begin
       hold.Add (thisprop);
     end;
   end;
  Q.free;
  o.free;
  s.addobject (objectclass.classname,hold) ;
  end;
end;


procedure ClassPropsList (Objectclass:TClass; s:TStrings);
var hold:Tstringlist; max,i:integer;
begin
  hold:=Tstringlist.create;
  BuildHierarchypropslist(Objectclass.classname,ObjectClass,s);
  max:= hold.count-1;
 { for i:=0 to max do
   begin
    if hold[i]=Objectclass.classname then s.text:=tstringlist(hold.objects[i]).text;
   end;}
  hold.free;
end;

function TQODB.FindClass (classname:String):TQODBClassDefinition;
var found:boolean;i,idx,max:integer; hold:TQODBCLassDefinition;uclassname:String;
begin
   max:=fclasses.count-1;
   found:=false;
   i:=0;
   idx:=-1;
   uclassname:=Uppercase(classname);
   result:=nil;
   while ((i<=max) and (not(found))) do
    begin
      hold:=TQODBCLassDefinition(fclasses.items[i]);
      if UpperCase(hold.name)=uclassname then
        begin
           idx:=i;
           result:=TQODBCLassDefinition(fclasses.items[i]);
        end;
      inc(i);
    end;
end;

procedure TQODB.RegisterClass (RegisteredClass:TClass;Prototype:TObject);
var Map,ThisClass:tstringlist; i,max,j,jmax:integer;ThisDef:TQODBClassDefinition; thisprop,thistype:String;
    Q:TRTTIENabler; anc:TClass;
begin
if prototype is TPersistent then
  if getclass(RegisteredClass.classname)=nil then
    classes.RegisterClass(TPersistentClass(registeredclass));
    
  Map:=tstringlist.create;
  {Creates basic class structures}
  ClassPropsList(RegisteredClass,Map);
  {Build Class Definitions}
  max:=Map.count-1;
  for i := 0 to max do
  if GetClass(Map[i])<>TObject then
   begin
     ThisClass:=TStringlist(Map.objects[i]);
     ThisDef:=TQODBClassDefinition(fclasses.add);
     ThisDef.ObjectClass := GetCLass(Map[i]);
     anc:=ThisDef.ObjectClass.ClassParent ;
     ThisDef.Ancestor := anc.ClassName ;
     thisdef.Name  := Map[i];
     jmax:=ThisClass.count-1;
     for j := 0 to jmax do
      begin
        thisprop:=thisclass.names[j];
        thistype:=ThisClass.values[thisprop];
        if thistype[1]='*' then
         begin
         delete(thistype,1,1);
          if FindCLass(thistype)=nil then  {register those sub objects!}
           if ProtoType is TPersistent then
           begin
             Q:=TRTTIEnabler.create;
             Q.rttiobject:=TPersistent(ProtoType);
             RegisterClass(GetClass(Thistype),Q.ChildObject (thisprop));
             Q.free;
           end;
         ThisDef.AddProperty(thisprop,thistype,true);
         end
         else
         ThisDef.AddProperty(thisprop,thistype,false);
      end;
   end;
  Map.free;
end;


function TQODB.FindObject(OID:TObjectID):TObject; {get obj ref from OID}
var found:boolean;i,idx,max:integer;
begin
  {fobjects keeps track of all the objects we pull from and put to the db, so that
   a reference to an object can be re-connected later... so that if 3 "invoice" objects
   point to one "customer", that the local memory will also do the same (instead of
   creating 3 separate customer objects)}
   {A hash would be good, but I'll use stringlist and objects}
   max:=fobjects.count-1;
   found:=false;
   i:=0;
   idx:=-1;
   result:=nil;
   while ((i<=max) and (not(found))) do
    begin
      if fobjects[i]=OID then
        begin
           idx:=i;
           result:=fobjects.objects[idx];
        end;
      inc(i);
    end;
end;

function TQODB.FindOID (Obj:TObject):TObjectID; {get objectID from obj ref}
var found:boolean;i,idx,max:integer;
begin
  {fobjects keeps track of all the objects we pull from and put to the db, so that
   a reference to an object can be re-connected later... so that if 3 "invoice" objects
   point to one "customer", that the local memory will also do the same (instead of
   creating 3 separate customer objects)}
   {A hash would be good, but I'll use stringlist and objects}
   max:=fobjects.count-1;
   found:=false;
   i:=0;
   idx:=-1;
   while ((i<=max) and (not(found))) do
    begin
      if fobjects.objects[idx]=Obj then
        begin
           idx:=i;
           result:=fobjects[idx];
        end;
      inc(i);
    end;
end;

constructor TQODBClassDefinition.create(Collection:TCollection);
begin
 inherited create(COllection);
 fprops:=TCollection.create(TQODBProperty);
end;

destructor  TQODBClassDefinition.destroy;
begin
 try
   fprops.free;
 finally
   inherited destroy;
 end;
end;

function TQODBClassDefinition.AddProperty(Name,ObjectType:String;IsObject:boolean):TQODBProperty;
begin
  result:=TQODBProperty(fprops.add);
  result.name:=name;
  result.propertytype:=ObjectType;
  result.isobject:=isobject;
end;

end.
